# dj
from django.utils.functional import cached_property

# internal
from .backends.base import BaseBackend
from .errors import PaymentBackendDoesNotExistError


class PayManager(object):
    """PayManager"""

    def __init__(self) -> None:
        self._configs = {
            "zarinpal": {
                "currency": "",
                "merchant_id": "",
                "callback_view_name": "",
            },
        }

    @property
    def zarinpal_currency(self) -> str:
        return self._configs["zarinpal"]["currency"]

    @zarinpal_currency.setter
    def zarinpal_currency(self, value: str) -> None:
        self._configs["zarinpal"]["currency"] = value

    @property
    def zarinpal_merchant_id(self) -> str:
        return self._configs["zarinpal"]["merchant_id"]

    @zarinpal_merchant_id.setter
    def zarinpal_merchant_id(self, value: str) -> None:
        self._configs["zarinpal"]["merchant_id"] = value

    @property
    def zarinpal_callback_view_name(self):
        return self._configs["zarinpal"]["callback_view_name"]

    @zarinpal_callback_view_name.setter
    def zarinpal_callback_view_name(self, value: str) -> None:
        self._configs["zarinpal"]["callback_view_name"] = value

    @cached_property
    def backends(self) -> list:
        from .backends import BACKENDS

        return BACKENDS

    @cached_property
    def backends_as_choices(self):
        return ((backend.identifier, backend.label) for backend in self.backends)

    def get_backend(self, identifier: str, config: dict | None = None) -> BaseBackend:
        for backend in self.backends:
            if identifier == backend.identifier:
                # get manager_config by given identifier
                manager_config = self._configs.get(identifier, {})
                # update manager_config by given config if any
                if config:
                    manager_config.update(config)
                return backend(manager_config)
        raise PaymentBackendDoesNotExistError



# dj
from django.contrib import admin

# internal
from .models import Bill


@admin.register(Bill)
class BillAdmin(admin.ModelAdmin):
    """Bill Admin"""

    model = Bill

    list_display = ["__str__", "amount", "created_at", "verified"]
    list_filter = ["backend", "verified", "created_at"]



# standard
from typing import Any

# internal
from ..models import Bill
from .base import BaseBackend
from ..errors import PaymentError


class PayOnDelivery(BaseBackend):
    """Pay On Delivery"""

    identifier = "pay-on-delivery"
    label = "Pay on delivery"

    def pay(self, amount: int, **extra: Any) -> Bill:
        return Bill.objects.create(backend=self.identifier, amount=amount, extra=extra)

    def verify(self, bill: Bill, **kwargs: Any) -> Bill:
        # check for backend
        if bill.backend != self.identifier:
            raise PaymentError("Invalid bill.")
        # check verified status
        if bill.verified:
            raise PaymentError("Invalid bill.")
        # verify and return bill
        bill.verified = True
        bill.save(update_fields=["verified"])
        return bill



# standard
from typing import Any

# internal
from ..models import Bill
from ..errors import PaymentError


class BaseBackend(object):
    """Base Backend"""

    identifier = "base"
    label = "Base"

    def __init__(self, config: dict | None = None) -> None:
        if config is None:
            config = {}
        self._config = self._validate_config(config)

    def _validate_config(self, config: dict) -> dict:  # noqa
        return config

    def _get_config(self, name: str, default: Any = None) -> Any:
        return self._config.get(name, default)

    def error(self, message: str) -> None:
        raise PaymentError(message)

    def pay(self, amount: int, **extra: Any) -> Bill:
        raise NotImplementedError

    def verify(self, bill: Bill, **kwargs: Any) -> Bill:
        raise NotImplementedError

    def __str__(self):
        return self.label

    def __repr__(self):
        return f"Backend(identifier={self.identifier}, label={self.label})"



class BasePaymentError(Exception):
    """Base Payment Error"""


class PaymentError(BasePaymentError):
    """Payment Error"""


class PaymentImproperlyConfiguredError(PaymentError):
    """Payment Improperly Configured Error"""


class PaymentBackendDoesNotExistError(PaymentError):
    """PaymentBackend Does Not Exist Error"""
